/*
 * mod_loopback.c for Apache 2.0
 *
 * Copyright 2001, 2003 by Anthony Howe.  All rights reserved.
 *
 *
 * LICENSE
 * -------
 *
 * This source distribution is made freely available and there is no charge
 * for its use, provided you retain this notice, disclaimers, author's
 * copyright, and credits.
 *
 *
 * DISCLAIMER
 * ----------
 *
 * THIS SOFTWARE IS PROVIDE "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO WAY SHALL THE
 * AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
 * OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/***********************************************************************
 *** No configuration below this point.
 ***********************************************************************/

#define MODULE			"mod_loopback"
#define AUTHOR			"achowe@snert.com"
#define VERSION			"2.1"
#define NDEBUG			1

#include "httpd.h"
#include "http_core.h"
#include "http_config.h"
#include "http_log.h"
#include "http_protocol.h"
#include "http_request.h"

#if defined(__unix__)
#include "unixd.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <time.h>

module AP_MODULE_DECLARE_DATA loopback_module;

static int loopback_log = 0;
static const char *loopback_log_file = "logs/loopback.log";

struct watchEntry {
	char *key;
	unsigned long ifInOctets;
	unsigned long ifOutOctets;
	unsigned long ifRequests;
	unsigned long ifDocuments;
	short         ifActive;
	double	      ifOutRate;
	unsigned long periodOctets;
	unsigned long periodMarker;
};

/***********************************************************************
 *** Support Routines
 ***********************************************************************/

static int
ml_istrue(const char *s)
{
	if (ap_strcasecmp_match(s, "enable") == 0)	/* disable */
		return 1;
	if (ap_strcasecmp_match(s, "true") == 0)	/* false */
		return 1;
	if (ap_strcasecmp_match(s, "yes") == 0)		/* no */
		return 1;
	if (ap_strcasecmp_match(s, "set") == 0)		/* reset */
		return 1;
	if (ap_strcasecmp_match(s, "ok") == 0)		/* bogus */
		return 1;
	if (ap_strcasecmp_match(s, "on") == 0)		/* off */
		return 1;
	if (ap_strcasecmp_match(s, "1") == 0)		/* 0 */
		return 1;

	return 0;
}

/*
 * Return a date string of the form "YYYY-MM-DD hh:mm:ss" representing
 * the current time in GMT.
 */
static char *
ml_gmt(apr_pool_t *p)
{
	int rc;
	time_t now;
	char *ascgmt;
	struct tm *gmt;

	now = time((time_t *) 0);

	if ((gmt = gmtime(&now)) == (struct tm *) 0)
		return (char *) 0;

	if ((ascgmt = apr_palloc(p, 21)) != (char *) 0)
		rc = strftime(ascgmt, 20, "%Y-%m-%d %H:%M:%S", gmt);

	return ascgmt;
}

/***********************************************************************
 ***  Create, Init, and Merge Routines
 ***********************************************************************/

static int
loopbackPostConfig(apr_pool_t *pconf, apr_pool_t *plog, apr_pool_t *ptemp, server_rec *s)
{
	int rc;
	apr_file_t *fp;

        ap_add_version_component(pconf, MODULE "/" VERSION);

        loopback_log_file = ap_server_root_relative(pconf, loopback_log_file);

	/* Create the log file in the logs directory with the
	 * child process' ownership.
	 */
	rc = apr_file_open(
		&fp, loopback_log_file,
		APR_CREATE|APR_WRITE|APR_APPEND|APR_BUFFERED,
		APR_UREAD|APR_UWRITE, ptemp
	);
	if (rc == APR_SUCCESS) {
#if defined(__unix__)
		(void) chown(loopback_log_file, unixd_config.user_id, -1);
#endif
		(void) apr_file_close(fp);
	}

        return OK;
}

/***********************************************************************
 ***  Handlers
 ***********************************************************************/

static int
headers(void *data, const char *key, const char *value)
{
	ap_rprintf((request_rec *) data, "%s: %s\r\n", key, value);

	return 1;
}

static int
logHeaders(void *data, const char *key, const char *value)
{
	(void) apr_file_printf((apr_file_t *) data, "%s: %s\n", key, value);

	return 1;
}

static apr_file_t *
logStart(request_rec *r)
{
	int rc;
	apr_file_t *fp;

	rc = apr_file_open(
		&fp, loopback_log_file,
		APR_CREATE|APR_WRITE|APR_APPEND|APR_BUFFERED,
		APR_UREAD|APR_UWRITE, r->pool
	);
	if (rc != APR_SUCCESS)
		return (apr_file_t *) 0;

	rc = apr_file_lock(fp, APR_FLOCK_EXCLUSIVE);
	if (rc != APR_SUCCESS) {
		(void) apr_file_close(fp);
		return (apr_file_t *) 0;
	}

	(void) apr_file_printf(fp, "---- %s %s\n", r->connection->remote_ip, ml_gmt(r->pool));
	(void) apr_file_printf(fp, "%s\n", r->the_request);
	apr_table_do(logHeaders, fp, r->headers_in, (char *) 0);
	(void) apr_file_printf(fp, "\n");

	return fp;
}

static void
loopbackDumpWatch(request_rec *r, apr_file_t *fp, const char *key)
{
	struct watchEntry *entry;

	entry = (struct watchEntry *) apr_table_get(r->notes, key);
	if (entry == (struct watchEntry *) 0)
		return;

	ap_rprintf(r, "\n[%s]\n", key);
	ap_rprintf(r, "key=%s\n", entry->key);
	ap_rprintf(r, "ifInOctets=%lu\n", entry->ifInOctets);
	ap_rprintf(r, "ifOutOctets=%lu\n", entry->ifOutOctets);
	ap_rprintf(r, "ifRequests=%lu\n", entry->ifRequests);
	ap_rprintf(r, "ifDocuments=%lu\n", entry->ifDocuments);
}

/*
 * This content handler returns a copy of of the request sent similar
 * to what the TRACE method would do.  Short of creating a special
 * HTML form that used the TRACE method, its not possible from Netscape
 * or IE to send a TRACE.
 *
 * Allows the URL used to perform a request loop-back:
 *
 *	http://www.snert.com/loop-back
 *
 *	<Location /loop-back>
 *	SetHandler loop-back
 *	</Location>
 */
static int
loopbackHandler(request_rec *r)
{
	int rc;
	char *buf;
	apr_size_t n;
	apr_file_t *log;

	if (r->handler == (char *) 0 || ap_strcmp_match(r->handler, "loop-back") != 0)
		return DECLINED;

	ap_set_content_type(r, "text/plain");
#ifdef WHAT_IS_NEW_API
	ap_send_http_header(r);
#endif

#ifdef DONT_IGNORE_HEAD
	if (r->header_only)
		return OK;
#endif

	log = loopback_log ? logStart(r) : (apr_file_t *) 0;

	ap_rprintf(r, "%s\r\n", r->the_request);
	apr_table_do(headers, r, r->headers_in, (char *) 0);
	ap_rprintf(r, "\r\n");

	if (r->method_number & (M_POST | M_PUT)) {
		if ((rc = ap_setup_client_block(r, REQUEST_CHUNKED_DECHUNK)) != OK)
			return rc;

		if (ap_should_client_block(r)) {
			buf = apr_palloc(r->pool, BUFSIZ);
#ifdef WHAT_IS_NEW_API
			ap_hard_timeout("loopback", r);
#endif
			while (0 < (n = ap_get_client_block(r, buf, BUFSIZ))) {
#ifdef WHAT_IS_NEW_API
				ap_reset_timeout(r);
#endif
				ap_rwrite(buf, n, r);

				if (log != (apr_file_t *) 0)
					(void) apr_file_write(log, buf, &n);
			}

#ifdef WHAT_IS_NEW_API
			ap_kill_timeout(r);
#endif
		}
	}

	loopbackDumpWatch(r, log, "watch-server-by-port");
	loopbackDumpWatch(r, log, "watch-server");
	loopbackDumpWatch(r, log, "watch-virtual-host-by-port");
	loopbackDumpWatch(r, log, "watch-virtual-host");
	loopbackDumpWatch(r, log, "watch-file-owner-by-port");
	loopbackDumpWatch(r, log, "watch-file-owner");
	loopbackDumpWatch(r, log, "watch-remote-ip-by-port");
	loopbackDumpWatch(r, log, "watch-remote-ip");
	loopbackDumpWatch(r, log, "watch-location");

	if (log != (apr_file_t *) 0) {
		(void) apr_file_printf(log, "\n");
		(void) apr_file_close(log);
	}

	return OK;
}

static void
loopbackHooks(apr_pool_t *p)
{
	ap_hook_post_config(loopbackPostConfig, NULL, NULL, APR_HOOK_MIDDLE);
	ap_hook_handler(loopbackHandler, NULL, NULL, APR_HOOK_MIDDLE);
}

/***********************************************************************
 ***  Configuration Directives
 ***********************************************************************/

/*
 * LoopbackLog {boolean}
 *
 * Context: server
 */
static const char *
LoopbackLog(cmd_parms *cmd, void *dconf, const char *arg)
{
	loopback_log = ml_istrue(arg);

	return (const char *) 0;
}

/*
 * LoopbackLogFile {file}
 *
 * Context: server
 */
static const char *
LoopbackLogFile(cmd_parms *cmd, void *dconf, const char *arg)
{
	/* Server relative or absolute path. */
	loopback_log_file = arg;

	return (const char *) 0;
}

static command_rec loopbackCommands[] = {
	AP_INIT_TAKE1(
		"LoopbackLog", LoopbackLog, NULL, RSRC_CONF,
		"Set true if all loopback requests should be logged."
	),
	AP_INIT_TAKE1(
		"LoopbackLogFile", LoopbackLogFile, NULL, RSRC_CONF,
		"Save the contents of the request to a file."
	),
	{ NULL }
};

/***********************************************************************
 ***  Module Definition Block
 ***********************************************************************/

module AP_MODULE_DECLARE_DATA loopback_module = {
	STANDARD20_MODULE_STUFF,
	NULL,			/* create per-dir    config structures */
	NULL,			/* merge  per-dir    config structures */
	NULL,			/* create per-server config structures */
	NULL,			/* merge  per-server config structures */
	loopbackCommands,	/* apr_table_t of config file commands       */
	loopbackHooks		/* register hooks */
};

